/*
 * HomeXMLExporter.java 
 *
 * Copyright (c) 2015 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.exporthtml5;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.media.j3d.BranchGroup;

import com.eteks.sweethome3d.j3d.ModelManager;
import com.eteks.sweethome3d.j3d.OBJWriter;
import com.eteks.sweethome3d.model.BackgroundImage;
import com.eteks.sweethome3d.model.Baseboard;
import com.eteks.sweethome3d.model.Camera;
import com.eteks.sweethome3d.model.Compass;
import com.eteks.sweethome3d.model.Content;
import com.eteks.sweethome3d.model.DimensionLine;
import com.eteks.sweethome3d.model.Home;
import com.eteks.sweethome3d.model.HomeDoorOrWindow;
import com.eteks.sweethome3d.model.HomeEnvironment;
import com.eteks.sweethome3d.model.HomeFurnitureGroup;
import com.eteks.sweethome3d.model.HomeLight;
import com.eteks.sweethome3d.model.HomeMaterial;
import com.eteks.sweethome3d.model.HomePieceOfFurniture;
import com.eteks.sweethome3d.model.HomePrint;
import com.eteks.sweethome3d.model.HomeTexture;
import com.eteks.sweethome3d.model.Label;
import com.eteks.sweethome3d.model.Level;
import com.eteks.sweethome3d.model.LightSource;
import com.eteks.sweethome3d.model.ObserverCamera;
import com.eteks.sweethome3d.model.Polyline;
import com.eteks.sweethome3d.model.Room;
import com.eteks.sweethome3d.model.Sash;
import com.eteks.sweethome3d.model.TextStyle;
import com.eteks.sweethome3d.model.Wall;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.tools.URLContent;

/**
 * Exporter for home instances.
 */
public class HomeXMLExporter extends ObjectXMLExporter<Home> {
  private String             homeName;
  private String             homeStructure;
  private int                flags;
  private HashSet<Content>   referencedContents;
  private Map<Level, String> levelIds;
  private Map<Wall, String>  wallIds;

  public HomeXMLExporter(Home home, String homeName, String homeStructure, int flags) {
    this.homeName = homeName;
    this.homeStructure = homeStructure;
    this.flags = flags;
    this.referencedContents = new HashSet<Content>();
    // Create level IDs
    int levelIndex = 0;
    this.levelIds = new HashMap<Level, String>();
    for (Level level : home.getLevels()) {
      String levelId = "level" + levelIndex++;
      this.levelIds.put(level, levelId);
    }
    // Create wall IDs
    int wallIndex = 0;
    this.wallIds = new HashMap<Wall, String>();
    for (Wall wall : home.getWalls()) {
      this.wallIds.put(wall, "wall" + wallIndex++);
    }
  }
  
  protected int getFlags() {
    return this.flags;
  }

  public HashSet<Content> getReferencedContents() {
    return this.referencedContents;
  }
  
  protected String getId(Object object) {
    if (object == null) {
      return null;
    } else if (object instanceof Level) {
      return this.levelIds.get(object);
    } else if (object instanceof Wall) {
      return this.wallIds.get(object);
    } else {
      throw new IllegalArgumentException("No Id provided for object of class " + object.getClass().getName());
    }
  }
  
  @Override
  protected void writeAttributes(XMLWriter writer, Home home) throws IOException {
    writer.writeAttribute("version", String.valueOf(home.getVersion()));
    if (getFlags() != 0) {
      writer.writeAttribute("exportFlags", String.valueOf(getFlags()));
    }
    if (home.getName() != null) {
      writer.writeAttribute("name", this.homeName);
    }
    if (this.homeStructure != null) {
      writer.writeAttribute("structure", this.homeStructure, null);
    }
    writer.writeAttribute("camera", home.getCamera() == home.getObserverCamera() ? "observerCamera" : "topCamera");
    writer.writeAttribute("selectedLevel", getId(home.getSelectedLevel()), null);
    if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
      writer.writeFloatAttribute("wallHeight", home.getWallHeight());
      writer.writeBooleanAttribute("basePlanLocked", home.isBasePlanLocked(), false);
      if (home.getFurnitureSortedProperty() != null) {
        writer.writeAttribute("furnitureSortedProperty", home.getFurnitureSortedProperty().name());
      }
      writer.writeBooleanAttribute("furnitureDescendingSorted", home.isFurnitureDescendingSorted(), false);
    }
  }
  
  @Override
  protected void writeChildren(XMLWriter writer, Home home) throws IOException {
    if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
      // Export furniture visible properties
      for (HomePieceOfFurniture.SortableProperty property : home.getFurnitureVisibleProperties()) {
        writer.writeStartElement("furnitureVisibleProperty");
        writer.writeAttribute("name", property.name());
        writer.writeEndElement();
      }
      // Export properties
      for (String propertyName : home.getPropertyNames()) {
        exportProperty(writer, propertyName, home.getProperty(propertyName));
      }
    }
    // Export environment, compass and cameras
    exportEnvironment(writer, home.getEnvironment());
    exportBackgroundImage(writer, home.getBackgroundImage());
    exportPrint(writer, home.getPrint());
    exportCompass(writer, home.getCompass());
    exportCamera(writer, home.getObserverCamera(), "observerCamera");
    exportCamera(writer, home.getTopCamera(), "topCamera");
    for (Camera camera : home.getStoredCameras()) {
      exportCamera(writer, camera, "storedCamera");
    }
    // Write Level elements
    for (Level level : home.getLevels()) {
      exportLevel(writer, level);
    }
    // Write furniture elements
    for (HomePieceOfFurniture piece : home.getFurniture()) {
      exportPieceOfFurniture(writer, piece);
    }
    if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
      for (Wall wall : home.getWalls()) {
        exportWall(writer, wall);
      }
      for (Room room : home.getRooms()) {
        exportRoom(writer, room);
      }
      for (Polyline polyline : home.getPolylines()) {
        exportPolyline(writer, polyline);
      }
      for (DimensionLine dimensionLine : home.getDimensionLines()) {
        exportDimensionLine(writer, dimensionLine);
      }
      for (Label label : home.getLabels()) {
        exportLabel(writer, label);
      }
    }
  }

  protected void exportEnvironment(XMLWriter writer, HomeEnvironment environment) throws IOException {
    new ObjectXMLExporter<HomeEnvironment>() {
        @Override
        protected void writeAttributes(XMLWriter writer, HomeEnvironment environment) throws IOException {
          writer.writeColorAttribute("groundColor", environment.getGroundColor());
          writer.writeColorAttribute("skyColor", environment.getSkyColor());
          writer.writeColorAttribute("lightColor", environment.getLightColor());
          writer.writeFloatAttribute("wallsAlpha", environment.getWallsAlpha(), 1);
          writer.writeBooleanAttribute("allLevelsVisible", environment.isAllLevelsVisible(), false);
          writer.writeBooleanAttribute("observerCameraElevationAdjusted", environment.isObserverCameraElevationAdjusted(), true);
          if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
            writer.writeColorAttribute("ceillingLightColor", environment.getCeillingLightColor());
            writer.writeAttribute("drawingMode", environment.getDrawingMode().name(), HomeEnvironment.DrawingMode.FILL.name());
            writer.writeFloatAttribute("subpartSizeUnderLight", environment.getSubpartSizeUnderLight(), 0);
            writer.writeIntegerAttribute("photoWidth", environment.getPhotoWidth());
            writer.writeIntegerAttribute("photoHeight", environment.getPhotoHeight());
            writer.writeAttribute("photoAspectRatio", environment.getPhotoAspectRatio().name());
            writer.writeIntegerAttribute("photoQuality", environment.getPhotoQuality());
            writer.writeIntegerAttribute("videoWidth", environment.getVideoWidth());
            writer.writeIntegerAttribute("videoHeight", environment.getVideoHeight());
            writer.writeAttribute("videoAspectRatio", environment.getVideoAspectRatio().name());
            writer.writeIntegerAttribute("videoQuality", environment.getVideoQuality());
            writer.writeIntegerAttribute("videoFrameRate", environment.getVideoFrameRate());
          }
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, HomeEnvironment environment) throws IOException {
          if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0
              && !environment.getVideoCameraPath().isEmpty()) {
            for (Camera camera : environment.getVideoCameraPath()) {
              exportCamera(writer, camera, "cameraPath");
            }
          }
          exportTexture(writer, environment.getGroundTexture(), "groundTexture");
          exportTexture(writer, environment.getSkyTexture(), "skyTexture");
        }
      }.writeElement(writer, environment);
  }

  protected void exportBackgroundImage(XMLWriter writer, BackgroundImage backgroundImage) throws IOException {
    if (backgroundImage != null) {
      if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
        new ObjectXMLExporter<BackgroundImage>() {
            @Override
            protected void writeAttributes(XMLWriter writer, BackgroundImage backgroundImage) throws IOException {
              writer.writeAttribute("image", getExportedContentName(backgroundImage.getImage()), null);
              writer.writeFloatAttribute("scaleDistanceXStart", backgroundImage.getScaleDistanceXStart());
              writer.writeFloatAttribute("scaleDistanceYStart", backgroundImage.getScaleDistanceYStart());
              writer.writeFloatAttribute("scaleDistanceXEnd", backgroundImage.getScaleDistanceXEnd());
              writer.writeFloatAttribute("scaleDistanceYEnd", backgroundImage.getScaleDistanceYEnd());
              writer.writeFloatAttribute("scaleDistance", backgroundImage.getScaleDistance());
              writer.writeFloatAttribute("xOrigin", backgroundImage.getXOrigin(), 0);
              writer.writeFloatAttribute("yOrigin", backgroundImage.getYOrigin(), 0);
              writer.writeBooleanAttribute("visible", backgroundImage.isVisible(), true);
            }
          }.writeElement(writer, backgroundImage);
      }
    }
  }

  protected void exportPrint(XMLWriter writer, HomePrint print) throws IOException {
    if (print != null) {
      if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
        new ObjectXMLExporter<HomePrint>() {
            @Override
            protected void writeAttributes(XMLWriter writer, HomePrint print) throws IOException {
              writer.writeAttribute("headerFormat", print.getHeaderFormat(), null);
              writer.writeAttribute("footerFormat", print.getFooterFormat(), null);
              writer.writeFloatAttribute("planScale", print.getPlanScale());
              writer.writeFloatAttribute("paperWidth", print.getPaperWidth());
              writer.writeFloatAttribute("paperHeight", print.getPaperHeight());
              writer.writeFloatAttribute("paperLeftMargin", print.getPaperLeftMargin());
              writer.writeFloatAttribute("paperTopMargin", print.getPaperTopMargin());
              writer.writeFloatAttribute("paperRightMargin", print.getPaperRightMargin());
              writer.writeFloatAttribute("paperBottomMargin", print.getPaperBottomMargin());
              writer.writeAttribute("paperOrientation", print.getPaperOrientation().name());
            }
          }.writeElement(writer, print);
      }
    }
  }

  protected void exportCompass(XMLWriter writer, Compass compass) throws IOException {
    if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
      new ObjectXMLExporter<Compass>() {
          @Override
          protected void writeAttributes(XMLWriter writer, Compass compass) throws IOException {
            writer.writeFloatAttribute("x", compass.getX());
            writer.writeFloatAttribute("y", compass.getY());
            writer.writeFloatAttribute("diameter", compass.getDiameter());
            writer.writeFloatAttribute("northDirection", compass.getNorthDirection());
            writer.writeFloatAttribute("logitude", compass.getLongitude());
            writer.writeFloatAttribute("latitude", compass.getLatitude());
            writer.writeAttribute("timeZone", compass.getTimeZone());
          }
        }.writeElement(writer, compass);
    }
  }

  protected void exportCamera(XMLWriter writer, Camera camera, final String attributeName) throws IOException {
    if (camera != null) {
      new ObjectXMLExporter<Camera>() {
          @Override
          protected void writeAttributes(XMLWriter writer, Camera camera) throws IOException {
            writer.writeAttribute("attribute", attributeName, null);
            writer.writeAttribute("name", camera.getName(), null);
            writer.writeAttribute("lens", camera.getLens().name());
            writer.writeFloatAttribute("x", camera.getX());
            writer.writeFloatAttribute("y", camera.getY());
            writer.writeFloatAttribute("z", camera.getZ());
            writer.writeFloatAttribute("yaw", camera.getYaw());
            writer.writeFloatAttribute("pitch", camera.getPitch());
            writer.writeFloatAttribute("fieldOfView", camera.getFieldOfView());
            writer.writeLongAttribute("time", camera.getTime());
            if (camera instanceof ObserverCamera) {
              writer.writeBooleanAttribute("fixedSize", ((ObserverCamera)camera).isFixedSize(), false);
            }
          }
        }.writeElement(writer, camera);
    }
  }

  protected void exportLevel(XMLWriter writer, Level level) throws IOException {
    new ObjectXMLExporter<Level>() {
        @Override
        protected void writeAttributes(XMLWriter writer, Level level) throws IOException {
          writer.writeAttribute("id", getId(level));
          writer.writeAttribute("name", level.getName());
          writer.writeFloatAttribute("elevation", level.getElevation());
          writer.writeFloatAttribute("floorThickness", level.getFloorThickness());
          writer.writeFloatAttribute("height", level.getHeight());
          writer.writeIntegerAttribute("elevationIndex", level.getElevationIndex());
          writer.writeBooleanAttribute("visible", level.isVisible(), true);
          writer.writeBooleanAttribute("viewable", level.isViewable(), true);
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, Level level) throws IOException {
          if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
            exportBackgroundImage(writer, level.getBackgroundImage());
          }
        }
      }.writeElement(writer, level);
  }

  protected void exportPieceOfFurniture(XMLWriter writer, HomePieceOfFurniture piece) throws IOException {
    if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0
        || piece.isVisible()) {
      new PieceOfFurnitureExporter().writeElement(writer, piece);
    }
  }
  
  protected class PieceOfFurnitureExporter extends ObjectXMLExporter<HomePieceOfFurniture> { 
    @Override
    protected void writeAttributes(XMLWriter writer, HomePieceOfFurniture piece) throws IOException {
      if (piece.getLevel() != null) {
        writer.writeAttribute("level", getId(piece.getLevel()));        
      }
      writer.writeAttribute("name", piece.getName());        
      writer.writeAttribute("catalogId", piece.getCatalogId(), null);
      writer.writeAttribute("creator", piece.getCreator(), null);
      writer.writeAttribute("model", getModelExportedContentName(piece.getModel()), null);
      if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
        writer.writeAttribute("icon", getExportedContentName(piece.getIcon()), null);
        writer.writeAttribute("planIcon", getExportedContentName(piece.getPlanIcon()), null);
      }
      writer.writeFloatAttribute("x", piece.getX());
      writer.writeFloatAttribute("y", piece.getY());
      writer.writeFloatAttribute("elevation", piece.getElevation(), 0f);
      writer.writeFloatAttribute("angle", piece.getAngle(), 0f);
      writer.writeFloatAttribute("width", piece.getWidth());
      writer.writeFloatAttribute("depth", piece.getDepth());
      writer.writeFloatAttribute("height", piece.getHeight());
      writer.writeBooleanAttribute("backFaceShown", piece.isBackFaceShown(), false);
      writer.writeBooleanAttribute("modelMirrored", piece.isModelMirrored(), false);
      writer.writeBooleanAttribute("visible", piece.isVisible(), true);
      writer.writeColorAttribute("color", piece.getColor());
      if (piece.getShininess() != null) {
        writer.writeFloatAttribute("shininess", piece.getShininess());
      }
      float [][] modelRotation = piece.getModelRotation();
      String modelRotationString = 
          floatToString(modelRotation[0][0]) + " " + floatToString(modelRotation[0][1]) + " " + floatToString(modelRotation[0][2]) + " "
        + floatToString(modelRotation[1][0]) + " " + floatToString(modelRotation[1][1]) + " " + floatToString(modelRotation[1][2]) + " "
        + floatToString(modelRotation[2][0]) + " " + floatToString(modelRotation[2][1]) + " " + floatToString(modelRotation[2][2]);
      writer.writeAttribute("modelRotation", modelRotationString, "1 0 0 0 1 0 0 0 1");
      if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
        writer.writeAttribute("description", piece.getDescription(), null);        
        writer.writeAttribute("information", piece.getInformation(), null);        
        writer.writeAttribute("staircaseCutOutShape", piece.getStaircaseCutOutShape(), null);
        writer.writeFloatAttribute("dropOnTopElevation", piece.getDropOnTopElevation(), 1f);
        writer.writeBooleanAttribute("doorOrWindow", piece.isDoorOrWindow(), false);
        writer.writeBooleanAttribute("movable", piece.isMovable(), true);
        writer.writeBooleanAttribute("resizable", piece.isResizable(), true);
        writer.writeBooleanAttribute("deformable", piece.isDeformable(), true);
        writer.writeBooleanAttribute("texturable", piece.isTexturable(), true);
        writer.writeBooleanAttribute("nameVisible", piece.isNameVisible(), false);
        writer.writeFloatAttribute("nameAngle", piece.getNameAngle(), 0f);
        writer.writeFloatAttribute("nameXOffset", piece.getNameXOffset(), 0f);
        writer.writeFloatAttribute("nameYOffset", piece.getNameYOffset(), 0f);
        writer.writeBigDecimalAttribute("price", piece.getPrice());
        writer.writeBigDecimalAttribute("valueAddedTaxPercentage", piece.getValueAddedTaxPercentage());
        writer.writeAttribute("currency", piece.getCurrency(), null);
      }
    }
    
    @Override
    protected void writeChildren(XMLWriter writer, HomePieceOfFurniture piece) throws IOException {
      // Write subclass child elements 
      if (piece instanceof HomeFurnitureGroup) {
        for (HomePieceOfFurniture groupPiece : ((HomeFurnitureGroup)piece).getFurniture()) {
          exportPieceOfFurniture(writer, groupPiece);
        }
      } else if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
        if (piece instanceof HomeLight) {
          HomeLight light = (HomeLight)piece;
          writer.writeFloatAttribute("power", light.getPower());
          for (LightSource lightSource : light.getLightSources()) {
            writer.writeStartElement("lightSource");
            writer.writeFloatAttribute("diameter", lightSource.getDiameter());
            writer.writeFloatAttribute("x", lightSource.getX());
            writer.writeFloatAttribute("y", lightSource.getY());
            writer.writeFloatAttribute("z", lightSource.getZ());
            writer.writeColorAttribute("color", lightSource.getColor());
            writer.writeEndElement();
          }
        } else if (piece instanceof HomeDoorOrWindow) {
          HomeDoorOrWindow doorOrWindow = (HomeDoorOrWindow)piece;
          writer.writeFloatAttribute("wallThickness", doorOrWindow.getWallThickness());
          writer.writeFloatAttribute("wallDistance", doorOrWindow.getWallDistance());
          writer.writeAttribute("cutOutShape", doorOrWindow.getCutOutShape(), null);
          writer.writeBooleanAttribute("boundToWall", doorOrWindow.isBoundToWall(), true);
          for (Sash sash : doorOrWindow.getSashes()) {
            writer.writeStartElement("sash");
            writer.writeFloatAttribute("xAxis", sash.getXAxis());
            writer.writeFloatAttribute("yAxis", sash.getYAxis());
            writer.writeFloatAttribute("width", sash.getWidth());
            writer.writeFloatAttribute("startAngle", sash.getStartAngle());
            writer.writeFloatAttribute("endAngle", sash.getEndAngle());
            writer.writeEndElement();
          }
        }
      }
      
      // Write child elements 
      if ((getFlags() & HomeXMLFileRecorder.INCLUDE_ONLY_VIEWER_DATA) == 0) {
        exportTextStyle(writer, piece.getNameStyle(), "nameStyle");
      }
      exportTexture(writer, piece.getTexture(), null);
      if (piece.getModelMaterials() != null) {
        for (HomeMaterial material : piece.getModelMaterials()) {
          exportMaterial(writer, material);
        }
      }
    }
  }

  protected void exportMaterial(XMLWriter writer, HomeMaterial material) throws IOException {
    if (material != null) {
      new ObjectXMLExporter<HomeMaterial>() {
          @Override
          protected void writeAttributes(XMLWriter writer, HomeMaterial material) throws IOException {
            writer.writeAttribute("name", material.getName());
            writer.writeColorAttribute("color", material.getColor());
            if (material.getShininess() != null) {
              writer.writeFloatAttribute("shininess", material.getShininess());
            }
          }
          
          @Override
          protected void writeChildren(XMLWriter writer, HomeMaterial material) throws IOException {
            exportTexture(writer, material.getTexture(), null);
          }
        }.writeElement(writer, material);
    }
  }
  
  protected void exportWall(XMLWriter writer, Wall wall) throws IOException {
    new ObjectXMLExporter<Wall>() {
        @Override
        protected void writeAttributes(XMLWriter writer, Wall wall) throws IOException {
          writer.writeAttribute("id", getId(wall));
          if (wall.getLevel() != null) {
            writer.writeAttribute("level", getId(wall.getLevel()));        
          }
          if (wall.getWallAtStart() != null) {
            writer.writeAttribute("wallAtStart", getId(wall.getWallAtStart()));
          }
          if (wall.getWallAtEnd() != null) {
            writer.writeAttribute("wallAtEnd", getId(wall.getWallAtEnd()));
          }
          writer.writeFloatAttribute("xStart", wall.getXStart());
          writer.writeFloatAttribute("yStart", wall.getYStart());
          writer.writeFloatAttribute("xEnd", wall.getXEnd());
          writer.writeFloatAttribute("yEnd", wall.getYEnd());
          writer.writeFloatAttribute("height", wall.getHeight());
          writer.writeFloatAttribute("heightAtEnd", wall.getHeightAtEnd());
          writer.writeFloatAttribute("thickness", wall.getThickness());
          writer.writeFloatAttribute("arcExtent", wall.getArcExtent());
          if (wall.getPattern() != null) {
            writer.writeAttribute("pattern", wall.getPattern().getName());
          }
          writer.writeColorAttribute("topColor", wall.getTopColor());
          writer.writeColorAttribute("leftSideColor", wall.getLeftSideColor());
          writer.writeFloatAttribute("leftSideShininess", wall.getLeftSideShininess(), 0);
          writer.writeColorAttribute("rightSideColor", wall.getRightSideColor());
          writer.writeFloatAttribute("rightSideShininess", wall.getRightSideShininess(), 0);
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, Wall wall) throws IOException {
          exportTexture(writer, wall.getLeftSideTexture(), "leftSideTexture");
          exportTexture(writer, wall.getRightSideTexture(), "rightSideTexture");
          exportBaseboard(writer, wall.getLeftSideBaseboard(), "leftSideBaseboard");
          exportBaseboard(writer, wall.getRightSideBaseboard(), "rightSideBaseboard");
        }
      }.writeElement(writer, wall);
  }
  
  protected void exportRoom(XMLWriter writer, Room room) throws IOException {
    new ObjectXMLExporter<Room>() {
        @Override
        protected void writeAttributes(XMLWriter writer, Room room) throws IOException {
          if (room.getLevel() != null) {
            writer.writeAttribute("level", getId(room.getLevel()));        
          }
          writer.writeAttribute("name", room.getName(), null);
          writer.writeFloatAttribute("nameAngle", room.getNameAngle(), 0f);
          writer.writeFloatAttribute("nameXOffset", room.getNameXOffset(), 0f);
          writer.writeFloatAttribute("nameYOffset", room.getNameYOffset(), 0f);
          writer.writeBooleanAttribute("nameVisible", room.isAreaVisible(), false);
          writer.writeFloatAttribute("areaAngle", room.getAreaAngle(), 0f);
          writer.writeFloatAttribute("areaXOffset", room.getAreaXOffset(), 0f);
          writer.writeFloatAttribute("areaYOffset", room.getAreaYOffset(), 0f);
          writer.writeColorAttribute("floorColor", room.getFloorColor());
          writer.writeFloatAttribute("floorShininess", room.getFloorShininess(), 0);
          writer.writeColorAttribute("ceilingColor", room.getCeilingColor());
          writer.writeFloatAttribute("ceilingShininess", room.getCeilingShininess(), 0);
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, Room room) throws IOException {
          exportTextStyle(writer, room.getNameStyle(), "nameStyle");
          exportTextStyle(writer, room.getAreaStyle(), "areaStyle");
          exportTexture(writer, room.getFloorTexture(), "floorTexture");
          exportTexture(writer, room.getCeilingTexture(), "ceilingTexture");
          for (float [] point : room.getPoints()) {
            writer.writeStartElement("point");
            writer.writeFloatAttribute("x", point [0]);
            writer.writeFloatAttribute("y", point [1]);
            writer.writeEndElement();
          }
        }
      }.writeElement(writer, room);
  }
  
  protected void exportPolyline(XMLWriter writer, Polyline polyline) throws IOException {
    new ObjectXMLExporter<Polyline>() {
        @Override
        protected void writeAttributes(XMLWriter writer, Polyline polyline) throws IOException {
          if (polyline.getLevel() != null) {
            writer.writeAttribute("level", getId(polyline.getLevel()));        
          }
          writer.writeFloatAttribute("thickness", polyline.getThickness(), 0f);
          writer.writeAttribute("capStyle", polyline.getCapStyle().name());
          writer.writeAttribute("joinStyle", polyline.getJoinStyle().name());
          writer.writeAttribute("dashStyle", polyline.getDashStyle().name());
          writer.writeAttribute("startArrowStyle", polyline.getStartArrowStyle().name());
          writer.writeAttribute("endArrowStyle", polyline.getEndArrowStyle().name());
          writer.writeColorAttribute("color", polyline.getColor());
          writer.writeBooleanAttribute("closedPath", polyline.isClosedPath(), false);
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, Polyline polyline) throws IOException {
          for (float [] point : polyline.getPoints()) {
            writer.writeStartElement("point");
            writer.writeFloatAttribute("x", point [0]);
            writer.writeFloatAttribute("y", point [1]);
            writer.writeEndElement();
          }
        }
      }.writeElement(writer, polyline);
  }
  
  protected void exportDimensionLine(XMLWriter writer, DimensionLine dimensionLine) throws IOException {
    new ObjectXMLExporter<DimensionLine>() {
        @Override
        protected void writeAttributes(XMLWriter writer, DimensionLine dimensionLine) throws IOException {
          if (dimensionLine.getLevel() != null) {
            writer.writeAttribute("level", getId(dimensionLine.getLevel()));        
          }
          writer.writeFloatAttribute("xStart", dimensionLine.getXStart());
          writer.writeFloatAttribute("yStart", dimensionLine.getYStart());
          writer.writeFloatAttribute("xEnd", dimensionLine.getXEnd());
          writer.writeFloatAttribute("yEnd", dimensionLine.getYEnd());
          writer.writeFloatAttribute("offset", dimensionLine.getOffset());
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, DimensionLine dimensionLine) throws IOException {
          exportTextStyle(writer, dimensionLine.getLengthStyle(), "lengthStyle");
        }
      }.writeElement(writer, dimensionLine);
  }

  protected void exportLabel(XMLWriter writer, Label label) throws IOException {
    new ObjectXMLExporter<Label>() {
        @Override
        protected void writeAttributes(XMLWriter writer, Label label) throws IOException {
          if (label.getLevel() != null) {
            writer.writeAttribute("level", getId(label.getLevel()));        
          }
          writer.writeFloatAttribute("x", label.getX());
          writer.writeFloatAttribute("y", label.getY());
          writer.writeFloatAttribute("angle", label.getAngle(), 0);
          writer.writeFloatAttribute("elevation", label.getElevation(), 0);
          writer.writeFloatAttribute("pitch", label.getPitch());
          writer.writeColorAttribute("color", label.getColor());
        }
        
        @Override
        protected void writeChildren(XMLWriter writer, Label label) throws IOException {
          exportTextStyle(writer, label.getStyle(), null);
          // Write text in a child element
          writer.writeStartElement("Text");
          writer.writeText(label.getText());
          writer.writeEndElement();
        }
      }.writeElement(writer, label);
  }
  
  protected void exportTextStyle(XMLWriter writer, TextStyle textStyle, 
                                 final String attributeName) throws IOException {
    if (textStyle != null) {
      new ObjectXMLExporter<TextStyle>() {
          @Override
          protected void writeAttributes(XMLWriter writer, TextStyle textStyle) throws IOException {
            writer.writeAttribute("attribute", attributeName, null);
            writer.writeAttribute("fontName", textStyle.getFontName(), null);
            writer.writeFloatAttribute("size", textStyle.getFontSize());
            writer.writeBooleanAttribute("italic", textStyle.isItalic(), false);
            writer.writeBooleanAttribute("bold", textStyle.isBold(), false);
          }
        }.writeElement(writer, textStyle);
    }
  }

  protected void exportBaseboard(XMLWriter writer, Baseboard baseboard, 
                                 final String attributeName) throws IOException {
    if (baseboard != null) {
      new ObjectXMLExporter<Baseboard>() {
          @Override
          protected void writeAttributes(XMLWriter writer, Baseboard baseboard) throws IOException {
            writer.writeAttribute("attribute", attributeName, null);
            writer.writeFloatAttribute("thickness", baseboard.getThickness());
            writer.writeFloatAttribute("height", baseboard.getHeight());
            writer.writeColorAttribute("color", baseboard.getColor());
          }
          
          @Override
          protected void writeChildren(XMLWriter writer, Baseboard baseboard) throws IOException {
            exportTexture(writer, baseboard.getTexture(), null);
          }
        }.writeElement(writer, baseboard);
    }
  }
  
  protected void exportTexture(XMLWriter writer, HomeTexture texture, 
                               final String attributeName) throws IOException {
    if (texture != null) {
      new ObjectXMLExporter<HomeTexture>() {
          @Override
          protected void writeAttributes(XMLWriter writer, HomeTexture texture) throws IOException {
            writer.writeAttribute("attribute", attributeName, null);
            writer.writeAttribute("name", texture.getName());
            writer.writeAttribute("catalogId", texture.getCatalogId(), null);
            writer.writeFloatAttribute("width", texture.getWidth());
            writer.writeFloatAttribute("height", texture.getHeight());
            writer.writeFloatAttribute("angle", texture.getAngle(), 0f);
            writer.writeAttribute("image", getExportedContentName(texture.getImage()), null);
          }
          
          @Override
          protected void writeChildren(XMLWriter writer, HomeTexture texture) throws IOException {
          }
        }.writeElement(writer, texture);
    }
  }
  
  private void exportProperty(XMLWriter writer, String propertyName, String propertyValue) throws IOException {
    writer.writeStartElement("property");
    writer.writeAttribute("name", propertyName);
    writer.writeAttribute("value", propertyValue);
    writer.writeEndElement();
  }

  protected String getExportedContentName(Content content) {
    if (content == null) {
      return null;
    } else {
      getReferencedContents().add(content);
      return content instanceof URLContent && ((URLContent)content).isJAREntry()
          ? ((URLContent)content).getJAREntryName()
          : content.toString();
    }
  }

  protected String getModelExportedContentName(Content content) throws IOException {
    if ((getFlags() & HomeXMLFileRecorder.CONVERT_MODELS_TO_OBJ_FORMAT ) != 0 
        && content != null 
        && (content instanceof URLContent)
        && !((URLContent)content).getURL().toString().endsWith(".obj")) {
      String modelFileName = "model.obj";
      String entryContent = ((URLContent)content).getJAREntryName();
      int slashIndex = entryContent.indexOf('/');
      String exportedEntry;
      if (slashIndex == -1) {
        exportedEntry = entryContent + "/" + modelFileName;
      } else {
        exportedEntry = entryContent.substring(0, slashIndex + 1) + modelFileName;
      }

      if (!getReferencedContents().contains(new URLContent(new URL(((URLContent)content).getURL(), "/" + exportedEntry)))) {
        // Convert model to OBJ format
        BranchGroup model = ModelManager.getInstance().loadModel(content);
        File tempModelFile = OperatingSystem.createTemporaryFile("model", ".zip");
        OBJWriter.writeNodeInZIPFile(model, tempModelFile, 0, modelFileName, "Export for compatibility");
        URLContent updatedContent = new URLContent(new URL("jar:" + tempModelFile.toURI().toURL() + "!/" + modelFileName));
        URLContent exportedContent = new RedirectedURLContent(new URL(((URLContent)content).getURL(), "/" + exportedEntry),
            updatedContent);
        getReferencedContents().add(exportedContent);
      }
      
      return exportedEntry;
    } else {
      return getExportedContentName(content);
    }
  }

  /**
   * Returns the string value of the given float, except for -1.0, 1.0 or 0.0
   * where -1, 1 and 0 is returned.
   */
  private static String floatToString(float f) {
    if (Math.abs(f) < 1E-6) {
      return "0";
    } else if (Math.abs(f - 1f) < 1E-6) {
      return "1";
    } else if (Math.abs(f + 1f) < 1E-6) {
      return "-1";
    } else {
      return String.valueOf(f);
    }
  }
}